Unit MKComFos;  {Fossil Comm object unit}

{$I MKB.Def}

Interface

{
     MKComFos - Copyright 1993 by Mark May - MK Software
     You are free to use this code in your programs, however
     it may not be included in Source/TPU function libraries
     without my permission.

     Mythical Kingom Tech BBS (513)237-7737 HST/v32
     FidoNet: 1:110/290
     Rime: ->MYTHKING
     You may also reach me at maym@dmapub.dma.org
}


Uses MKComAbs,
  {$IFDEF WINDOWS}
  WinDos;
  {$ELSE}
  Dos;
  {$ENDIF}


Type FosComObj = Object(AbsComObj)
  {$IFDEF WINDOWS}
  Regs: TRegisters;
  {$ELSE}
  Regs: Registers;
  {$ENDIF}
  Constructor Init;
  Destructor Done; Virtual;
  Function OpenPort(PortNo: Byte; Baud: LongInt; Flow: Byte; NumBits: Byte;
    StopBits: Byte; Parity: Byte): Boolean; Virtual;
  Procedure ClosePort; Virtual;
  Procedure SetBPS(Rate: LongInt); Virtual;
  Function  GetBPS: LongInt; Virtual;
  Procedure LowerDTR; Virtual;
  Procedure RaiseDTR; Virtual;
  Function  GetInfo: String; Virtual;
  Function  OutReady: Boolean; Virtual;
  Function  OutEmpty: Boolean; Virtual;
  Function  InEmpty: Boolean; Virtual;
  Function  GetChar: Char; Virtual;
  Function  PeekChar: Char; Virtual;
  Procedure SendChar(Ch: Char); Virtual;
  Function  SendCharNW(Ch: Char): Boolean; Virtual;
  Procedure SetFlowCtl(FC: Byte); Virtual;
  Procedure StartBreak; Virtual;
  Procedure StopBreak; Virtual;
  Function  SendBlock(Var Dt; DSize: Word): Word; Virtual;
  Function  GetBlock(Var Dt; DSize: Word): Word; Virtual;
  Function  GetCarrier: Boolean; Virtual;
  Procedure FlushIn; Virtual;
  Procedure FlushOut; Virtual;
  End;

Type FosComPtr = ^FosComObj;

Function DetectFossil(FosPort: Word): Boolean;

Implementation

Type FossilInfo = Record
  fiSize: Word;
  fiSpec: Byte;
  fiRev: Byte;
  fiId: Pointer;
  fiInSize: Word;
  fiInFree: Word;
  fiOutSize: Word;
  fiOutFree: Word;
  fiSWidth: Word;
  fiSHeight: Word;
  fiBaudMask: Word;
  End;


Function DetectFossil(FosPort: Word): Boolean;
  Var
  {$IFDEF WINDOWS}
    Regs: TRegisters;
  {$ELSE}
    Regs: Registers;
  {$ENDIF}

  Begin
  Regs.ah := $04;
  Regs.dx := FosPort - 1;
  Regs.bx := 0;
  Intr($14, Regs);
  DetectFossil := Regs.ax = $1954;
  Regs.ah := $05;
  Regs.dx := FosPort - 1;
  Intr($14, Regs);
  End;


Constructor FosComObj.Init;
  Begin
  coBaud := 2400;
  coPort := 1;
  coFlow := 0;
  coParity := 0;
  coNumBits := 8;
  coStopBits := 1;
  End;


Destructor FosComObj.Done;
  Begin
  End;


Function FosComObj.OpenPort(PortNo: Byte; Baud: LongInt; Flow: Byte; NumBits: Byte;
  StopBits: Byte; Parity: Byte): Boolean;
  Begin
  coPort := PortNo;
  coBaud := Baud;
  coFlow := Flow;
  coNumBits := NumBits;
  coStopBits := StopBits;
  coParity := Parity;
  Regs.ah := $04;
  Regs.dx := coPort - 1;
  Regs.bx := 0;
  Intr($14, Regs);
  OpenPort := Regs.Ax = $1954;
  {set baud}
  {set flow}
  {set numbits, stop, and parity}
  End;


Procedure FosComObj.ClosePort;
  Begin
  Regs.ah := $05;
  Regs.dx := coPort - 1;
  Intr($14, Regs);
  End;


Procedure FosComObj.SetBPS(Rate: LongInt);
  Begin
  coBaud := Rate;
  Regs.ah := $00;
  Regs.dx := coPort - 1;
  Case Word(coBaud) of
    00300: Regs.al := 64;
    00600: Regs.al := 96;
    01200: Regs.al := 128;
    02400: Regs.al := 160;
    04800: Regs.al := 192;
    09600: Regs.al := 224;
    19200: Regs.al := 0;
    38400: Regs.al := 32;
    End;
  Inc(Regs.al, 3);
  Intr($14, Regs);
  End;


Function FosComObj.GetBPS: LongInt;
  Begin
  GetBPS := coBaud;
  End;


Procedure FosComObj.LowerDTR;
  Begin
  Regs.ah := $06;
  Regs.dx := coPort - 1;
  Regs.al := 0;
  Intr($14, Regs);
  End;


Procedure FosComObj.RaiseDTR;
  Begin
  Regs.ah := $06;
  Regs.dx := coPort - 1;
  Regs.al := 1;
  Intr($14, Regs);
  End;


Function FosComObj.GetInfo: String;
  Var
    Info: FossilInfo;
    TmpStr: String;

  Begin
  Regs.ah := $1b;
  Regs.cx := SizeOf(Info);
  Regs.es := Seg(Info);
  Regs.di := Ofs(Info);
  Regs.dx := coPort - 1;
  Intr($14, Regs);
  Move(Info.fiId^, TmpStr[1], 255);
  TmpStr[0] := #255;
  TmpStr[255] := #0;
  TmpStr[0] := chr(Pos(#0, TmpStr));
  GetInfo := TmpStr;
  End;


Function FosComObj.OutReady: Boolean;
  Begin
  Regs.ah := $03;
  Regs.dx := coPort - 1;
  Intr($14, Regs);
  OutReady := (Regs.ah and 32) <> 0;
  End;


Function FosComObj.OutEmpty: Boolean;
  Begin
  Regs.ah := $03;
  Regs.dx := coPort - 1;
  Intr($14, Regs);
  OutEmpty := (Regs.ah and 64) <> 0;
  End;


Function FosComObj.InEmpty: Boolean;
  Begin
  Regs.ah := $03;
  Regs.dx := coPort - 1;
  Intr($14, Regs);
  InEmpty := (Regs.ah and 1) = 0;
  End;


Function FosComObj.GetChar: Char;
  Begin
  Regs.ah := $02;
  Regs.dx := coPort - 1;
  Intr($14, Regs);
  GetChar := Chr(Regs.al);
  End;


Function FosComObj.PeekChar: Char;
  Begin
  Regs.ah := $0c;
  Regs.dx := coPort - 1;
  Intr($14, Regs);
  PeekChar := Chr(Regs.al);
  End;


Procedure FosComObj.SendChar(Ch: Char);
  Begin
  Regs.ah := $0b;
  Regs.dx := coPort - 1;
  Regs.al := Ord(Ch);
  Intr($14, Regs);
  End;


Function FosComObj.SendCharNW(Ch: Char): Boolean;
  Begin
  Regs.ah := $0b;
  Regs.dx := coPort - 1;
  Regs.al := Ord(ch);
  Intr($14, Regs);
  SendCharNW := (Regs.Ax <> 0);
  End;


Procedure FosComObj.SetFlowCtl(FC: Byte);
  Begin
  coFlow := FC;
  Regs.ah := $0f;
  Regs.dx := coPort - 1;
  Regs.al := FC;
  Intr($14, Regs);
  End;


Procedure FosComObj.StartBreak;
  Begin
  Regs.ah := $1a;
  Regs.dx := coPort - 1;
  Regs.al := 1;
  Intr($14, Regs);
  End;


Procedure FosComObj.StopBreak;
  Begin
  Regs.ah := $1a;
  Regs.dx := coPort - 1;
  Regs.al := 0;
  Intr($14, Regs);
  End;


Function FosComObj.SendBlock(Var Dt; DSize: Word): Word;
  Begin
  Regs.ah := $19;
  Regs.dx := coPort - 1;
  Regs.es := Seg(dt);
  Regs.di := Ofs(dt);
  Regs.dx := DSize;
  Intr($14, Regs);
  SendBlock := Regs.ax;
  End;


Function FosComObj.GetBlock(Var Dt; DSize: Word): Word;
  Begin
  Regs.ah := $18;
  Regs.dx := coPort - 1;
  Regs.es := Seg(dt);
  Regs.di := Ofs(dt);
  Regs.dx := DSize;
  Intr($14, Regs);
  GetBlock := Regs.ax;
  End;


Function FosComObj.GetCarrier: Boolean;
  Begin
  Regs.ah := $03;
  Regs.dx := coPort - 1;
  Intr($14, Regs);
  GetCarrier := (Regs.al and 128) <> 0;
  End;


Procedure FosComObj.FlushIn;
  Begin
  Regs.ah := $0a;
  Regs.dx := coPort - 1;
  Intr($14, Regs);
  End;


Procedure FosComObj.FlushOut;
  Begin
  Regs.ah := $09;
  Regs.dx := coPort - 1;
  Intr($14, Regs);
  End;


End.
